/****************************************************************************
**
** Copyright (C) 1991-2008 Heinz van Saanen
**
** This file is part of the the function viewer Zhu3D.
**
** This file may be used under the terms of the GNU General Public
** License version 2 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/


// Includes
#include <QApplication>
#include <QIcon>
#include <QMessageBox>
#include <QObject>
#include <QTranslator>
#include <QtGlobal>

#include "debug.h"
#include "property.h"
#include "mainwindow.h"
#include "glwidget.h"
#include "ligedit.h"
#include "matedit.h"
#include "entedit.h"
#include "aniedit.h"
#include "moredit.h"
#include "picedit.h"
#include "fntedit.h"
#include "demedit.h"
#include "funedit.h"
#include "speedit.h"
#include "diredit.h"
#include "usredit.h"
#include "legedit.h"
#include "sysinfo.h"


// Miscellaneous globals
QString appPath;		// Path where Zhu3D was started from
QString argStr;			// Holds arguments from bash or window-manager
bool gotArg=false;		// Where there any arguments?
float MHz=0.0f;			// Stays zero, when no supported hw-timer is found

// Denotes the number of real CPU-cores. The file property.cpp
// provides numCPUset to detect the number of user defined threads
int numCPU;


// Global widgets
MaiWindow *maiWin;		// Global because the system-font-stuff needs it
OGLWidget *oglWid;		// OpenGL viewer
usrWidget *usrEdi;		// User defined functions/constants editor
ligWidget *ligEdi;		// Light editor
matWidget *matEdi;		// Material editor
entWidget *entEdi;		// Entire editor
aniWidget *aniEdi;		// Animation editor
morWidget *morEdi;		// Morphing editor
picWidget *picEdi;		// Pic-dimension editor
fntWidget *fntEdi;		// Fontsize editor
demWidget *demEdi;		// Demo mode
funWidget *funEdi;		// Function/parameter editor
speWidget *speEdi;		// Special settings
dirWidget *dirEdi;		// Directory settings
legWidget *legEdi;		// Legends editor
#ifdef ZHUDEBUG
errWidget *errEdi;		// Error messages
#endif


// Localization-infos loading help-files and other stuff
// Used extern in modules fntedit, mainwindow and usredit
QString zhuLocale;
QTranslator zhuTranslator;


// Install own message handler
// Normal stderr iand Qt4-messages are redirected to error window
#ifdef ZHUDEBUG
void myMessageHandler(QtMsgType type, const char *msg) {
	switch (type) {
		case QtDebugMsg:
			PRINT((QString)"Debug: "+msg);
			break;
		case QtWarningMsg:
			PRINT((QString)"Warning: "+msg);
			break;
		case QtCriticalMsg:
			PRINT((QString)"Critical: "+msg);
			break;
		case QtFatalMsg:
			PRINT((QString)"Fatal: "+msg);
			break;
		abort();
	}
}
#endif


// Include stuff for detecting number of CPU cores
#if defined(Q_WS_MAC)
	#include <sys/types.h>
	#include <sys/param.h>
	#include <sys/sysctl.h>

#elif defined(Q_WS_WIN)
	#include <windows.h>

#elif defined(Q_OS_LINUX)
	#include <unistd.h>
#endif


// Returns number of CPU cores for misc. OS's
static int cpu_core_count() {
	int tmp;

	// Apple
	#if defined(Q_WS_MAC)
		int mib[2]={ CTL_HW, HW_NCPU };
		int cpus;
		size_t size=sizeof(cpus);
		if(0==sysctl(mib, 2, &cpus, &size, NULL, 0))
			tmp=cpus;

	// Windows
	#elif defined(Q_WS_WIN)
		SYSTEM_INFO si;
		GetSystemInfo(&si);
		tmp=si.dwNumberOfProcessors;

	// SGI/IRIX
	#elif defined(sgi)
		tmp=sysconf(_SC_NPROC_ONLN);

	// Linux/Unix
	#elif defined(Q_OS_LINUX)
		#if defined(_SC_NPROCESSORS_ONLN)
			tmp=sysconf(_SC_NPROCESSORS_ONLN);
		#endif
	#endif

    return ((tmp<1) || (tmp>MAXCPU)) ? 1 : tmp;
}


/* *************************************************** */
/* MAIN                                                */
/* *************************************************** */
int main(int argc, char *argv[]) {

	// Very basic initializations
	#ifdef ZHUDEBUG
	qInstallMsgHandler(myMessageHandler);
	#endif
	QApplication app(argc, argv);

	// Very first, maybe needed from others
	#ifdef ZHUDEBUG
	errEdi = new errWidget;
	errEdi->show();
	#endif

	// Detect the number of real CPU-cores
	// Default numCPUset=numCPU, what may be altered from fntEdi later
	numCPU = numCPUset = cpu_core_count();

	// Initialize logo and application-path
	QIcon appIcon(":/images/icon.png");
	app.setWindowIcon(appIcon);
	appPath=QCoreApplication::applicationDirPath();

	// General internationalization stuff
	QTranslator qtTranslator;
	qtTranslator.load("qt_"+zhuLocale);
	app.installTranslator(&qtTranslator);

	// Set Zhu3D languages
	zhuLocale=QLocale::system().name();
	zhuLocale.resize(zhuLocale.lastIndexOf('_'));
	zhuTranslator.load(MYLANDIR+"zhu3d_"+zhuLocale);
	app.installTranslator(&zhuTranslator);

	// Check arguments from command line
	if(argc==2) {
		argStr=argv[1];
		gotArg=true;
	}

	// Some more basic initializations
	initSysInfo();
	fpError::initParserQt();
	if(!initAllParsers()) {
		QMessageBox::critical(
					0,
					QObject::tr("Program start")+"\n",
					fpError::getQtMessage(11)
		);
		return -1;
	}

	// OpenGL defaults. Enable motion blur through accumulation buffer
	QGLFormat fmt;
	fmt.setAccum(true);
	fmt.setAlpha(true);
	if(!fmt.hasOpenGL()) {
		QMessageBox::critical(
					0,
					QObject::tr("Program start")+"\n",
					QObject::tr("Missing OpenGL! Terminating now")
		);
		return -1;
	}
	QGLFormat::setDefaultFormat(fmt);

	// Create globals
	oglWid = new OGLWidget;		// OpenGL viewer first, some inits done here!
	usrEdi = new usrWidget;		// User defined functions editor next! Needed from fntEdi later

	ligEdi = new ligWidget;		// Light editor
	matEdi = new matWidget;		// Material editor
	entEdi = new entWidget;		// Entire editor
	aniEdi = new aniWidget;		// Animation editor
	morEdi = new morWidget;		// Morphing editor
	picEdi = new picWidget;		// Pic-dimension editor
	funEdi = new funWidget;		// Function editor
	fntEdi = new fntWidget;		// Appeareance settings
	demEdi = new demWidget;		// Demo mode
	speEdi = new speWidget;		// Specials settings
	dirEdi = new dirWidget;		// Directory settings
	legEdi = new legWidget;		// Legends editor

	// Check clock frequency. Hardware dependent CPU timer has to be implemented in
	// ticks.h for this. If not, this does no harm and frequency is just "unknown"
	// For Windows this is disabled automatically because of bad timer-granularity
	#if defined(__i386__) || defined(__x86_64__) || \
		defined(__powerpc__) || \
		(defined(__mips__) && (defined(MIPSEL) || defined (__MIPSEL__)))
		MHz=aniEdi->getClock();
	#endif

	// Create and bring up the mainwindow finally
	maiWin = new MaiWindow;
	maiWin->show();

	return app.exec();
}
